"""Put user defined tasks in the plugins folder. You can start with
some customizations in this file which is included by default."""

from invoke import task

from ..config import (
    TEST_MESSAGE,
    SYS_TEST_MESSAGE,
)

@task
def hello(cx):
    """An example custom user task from the plugins."""
    print("Hello!")
    print(f"Message from config.py: {TEST_MESSAGE}")
    print(f"Message from sysconfig.py: {SYS_TEST_MESSAGE}")



"""Configuration managed by the system. Do not make changes you aren't
afraid of being overwritten and removed automatically."""

SYS_TEST_MESSAGE = "Hi! Jubeo taskset author here, please don't override these options " \
                   "unless you know what you are doing."



"""Specify which plugins to load"""

# import plugins:

from . import custom

# specify which plugins to install, the custom one is included by
# default to get users going
PLUGIN_MODULES = [
    custom,
]



"""User settings for a project."""

# load the system configuration. You can override them in this module,
# but beware it might break stuff
from .sysconfig import *

TEST_MESSAGE = "You can configure modules from this file"



import sys
from pathlib import Path
import pkgutil
import shutil
import os
import os.path as osp
import dataclasses as dc
from typing import (
    Tuple,
    Optional,
)

import toml
from hyperlink import URL
from git import Repo

JUBEO_PROJ_DIR = ".jubeo"

JUBEO_PROJ_CONF = "jubeo.toml"
"""The config file for individual projects."""

TASKSET_NAME = 'tasks'


@dc.dataclass
class ProjConfig():

    upstream_url: URL
    modules_source_url: URL
    modules: Tuple[str]
    taskset_dirname: str

def base_repo_path():

    return Path(sys.modules[__name__].__file__).parent / 'base_repo'

def config_template_path():

    return Path(sys.modules[__name__].__file__).parent / 'config_template'

def get_repo(url, cache_path):

    # if remote download (via git)
    if url.scheme == "git+https":

        # get all but the last '.git' at the end
        repo_name = url.path[-1].replace('.git', '')

        repo_cache_path = Path(cache_path) / repo_name

        print(f"Downloading {repo_name}")
        # remove the "git+" for the cloning
        git_repo = Repo.clone_from(
            url.replace(scheme="https", fragment="").to_text(),
            repo_cache_path)
        print("Done Downloading")

        git_repo_path = Path(git_repo.working_tree_dir)

        # resolve whether the source bundle is at the toplevel or
        # within
        if url.fragment == '':
            source_repo_path = git_repo_path

        else:
            source_repo_path = git_repo_path / url.fragment


    elif url.scheme == "https" or url.scheme == "http":
        raise NotImplementedError("Please use a local file, or the 'git+https' scheme")

    elif url.scheme == '' or url.scheme == 'file':

        print("Retrieving from local repository.")

        # if it is a rooted path add the root
        if url.rooted:
            origin_repo_path = Path('/' + '/'.join(url.path))
        else:
            origin_repo_path = Path('/'.join(url.path))


        # get all but the last '.git' at the end
        repo_name = url.path[-1]

        repo_cache_path = Path(cache_path) / repo_name

        # copy the repo to the cache since we will be manipulating it
        shutil.copytree(
            origin_repo_path,
            repo_cache_path,
        )

        source_repo_path = repo_cache_path / url.fragment

    else:
        raise ValueError(f"URL for upstream jubeo spec invalid: {url.to_text()}")

    return source_repo_path

def retrieve_upstream(
        url: URL,
        cache_path: Path,
        proj_config = None
):

    # TODO: use ProjConfig API, now we are just using a dict

    cach_path = Path(cache_path)
    repo_cache_path = cache_path / "specs"

    # get the spec repo
    upstream_repo_path = get_repo(url, repo_cache_path)

    # load the local config otherwise use the upstream's default
    # config
    if proj_config is None:
        proj_config = load_proj_config(upstream_repo_path)

    ## Retrieve external modules

    mod_source_url = URL.from_text(
        osp.expandvars(osp.expanduser(proj_config['modules']['source_url'])))

    # we support retrieving modules from a 3rd party source via URL
    modules = proj_config['modules']['modules']

    # then get the modules repo if any where requested
    if len(modules):

        mod_cache_path = cache_path / "modules"
        mod_repo_path = get_repo(mod_source_url, mod_cache_path)

        # then copy the modules to the source repo
        for module in modules:
            mod_fname = f"{module}.py"
            shutil.copyfile(
                mod_repo_path / mod_fname,
                upstream_repo_path / 'tasks' / 'modules' / mod_fname,
            )

    return upstream_repo_path


def load_proj_config(proj_dir):

    proj_config_path = Path(proj_dir) / JUBEO_PROJ_DIR / JUBEO_PROJ_CONF
    proj_config = toml.load(proj_config_path)

    return proj_config


def update_project(
        source,
        target,
        config,
        force=True
):

    # configuration folder
    source_config = Path(source) / JUBEO_PROJ_DIR
    target_config = Path(target) / JUBEO_PROJ_DIR

    update_config(
        source_config,
        target_config,
        force=force,
    )

    # taskset dir
    source_taskset = Path(source) / TASKSET_NAME
    target_taskset = Path(target) / config['taskset']['dirname']

    update_taskset(
        source_taskset,
        target_taskset,
        config,
        force=force,
    )


def update_config(
        source,
        target,
        force=True,
):
    """Update the configuration files for this project.

    Parameters
    ----------

    source : Path
        The configuration repository to copy from.

    target : Path
        The target configuration folder to update.

    """

    source = Path(source)
    target = Path(target)

    managed_files = (
        'requirements.in',
        'requirements.txt',
    )

    if (any([osp.exists(target / f)
             for f in managed_files])
        and not force):

        raise OSError("Project config exists, not overwriting")

    elif force:
        for f in managed_files:
            f = target / f
            if osp.isdir(f):
                print(f"Cleaning {f}")
                shutil.rmtree(f)
            elif osp.isfile(f):
                print(f"Cleaning {f}")
                os.remove(f)

    print("Updating .jubeo/requirements.in")
    shutil.copyfile(
        source / "requirements.in",
        target / "requirements.in"
    )

    print("Updating .jubeo/requirements.txt")
    shutil.copyfile(
        source / "requirements.txt",
        target / "requirements.txt"
    )


def update_taskset(
        source,
        target,
        config,
        force=True,
):
    """Update the taskset for this project.

    Parameters
    ----------

    source : Path
        The task-set repository to copy from.

    target : Path
        The target task-set to update.

    """

    source = Path(source)
    target = Path(target)

    managed_files = (
        '__init__.py',
        'sysconfig.py',
        'modules',
    )

    if (any([osp.exists(target / f)
             for f in managed_files])
        and not force):

        raise OSError("Project taskset file exists, not overwriting")

    elif force:
        for f in managed_files:
            f = target / f
            if osp.isdir(f):
                print(f"Cleaning {f}")
                shutil.rmtree(f)
            elif osp.isfile(f):
                print(f"Cleaning {f}")
                os.remove(f)

    # then get modules we need and replace the ones in this project
    # with them
    print("Updating tasks/sysconfig.py")
    shutil.copyfile(
        source / "sysconfig.py",
       target / "sysconfig.py"
    )

    print("Updating tasks/__init__.py")
    shutil.copyfile(
        source / "__init__.py",
        target / "__init__.py"
    )

    print("Updating tasks/modules")
    shutil.copytree(
        source / "modules",
        target / "modules",
    )

def init_project(
        source_repo,
        target_project,
        force=False,
):

    source_repo = Path(source_repo)
    target_project = Path(target_project)

    # load the configuration in the source repo
    proj_config = load_proj_config(source_repo)

    ## configuration directory

    configdir_source = source_repo / JUBEO_PROJ_DIR
    configdir_target = target_project / JUBEO_PROJ_DIR

    # make the config directory
    os.makedirs(configdir_target, exist_ok=True)

    # copy the config file: since the update config doesn't manage it
    configfile_source = configdir_source / JUBEO_PROJ_CONF
    configfile_target = configdir_target / JUBEO_PROJ_CONF

    if osp.exists(configfile_target):

        if not force:
            raise OSError("Configuration file exists, not overwriting")

    shutil.copyfile(
        configfile_source,
        configfile_target,
    )

    # not all of the taskset is taken care of by init so do that here
    taskset_source = source_repo / TASKSET_NAME
    taskset_target = target_project / proj_config['taskset']['dirname']

    if osp.exists(taskset_target) and force:
        shutil.rmtree(taskset_target)

    shutil.copytree(
        taskset_source,
        taskset_target,
        dirs_exist_ok=force,
        ignore=shutil.ignore_patterns("*~"),
    )

    # then just use update to copy everything else just to make sure
    update_project(
        source_repo,
        target_project,
        proj_config,
        force=True,
    )



from invoke import Collection, Task, task

import inspect


## Utilities

# these helper functions are for automatically listing all of the
# functions defined in the tasks module

def _is_mod_task(mod, func):
    return issubclass(type(func), Task) and inspect.getmodule(func) == mod

def _get_functions(mod):
    """get only the functions that aren't module functions and that
    aren't private (i.e. start with a '_')"""

    return {func.__name__ : func for func in mod.__dict__.values()
            if _is_mod_task(mod, func) }


## Namespace

# add all of the modules to the CLI
ns = Collection()

## Top-level

from . import toplevel
for func in _get_functions(toplevel).values():
    ns.add_task(func)

## Upstream

from .modules import MODULES as modules

for module in modules:
    ns.add_collection(module)

## Plugins

try:
    # import all the user defined stuff and override
    from .plugins import PLUGIN_MODULES as plugins

    for module in plugins:
        ns.add_collection(module)

except Exception as e:
    print("Loading plugins failed with error ignoring:")
    print(e)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import setup, find_packages

import itertools as it

# the basic needed requirements for a package
base_requirements = [
    'gitpython',
    'click',
    'hyperlink',
    'toml',
]

# extras requirements list

# SNIPPET: example extra requirement
# example_extra_requirements = ['requests']
# extras = [example_extra_requirements,]

# Add your extra requirements lists here:
extras = [
]

# combination of all the extras requirements
_all_requirements = [[base_requirements]] + extras
all_requirements = it.chain.from_iterable(_all_requirements)

setup(
    name='jubeo',
    version='0.0.3a0.dev0',
    author="Samuel D. Lotz",
    author_email="samuel.lotz@salotz.info",
    description="Small tool for deploying and updating metaproject tooling.",
    #long_description=open('README.org').read(),
    license="MIT",
    url="https://github.com/salotz/jubeo",
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3'
    ],
    # building/dev
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'tox'],

    # package
    packages=find_packages(where='src'),

    package_dir={'' : 'src'},

    # if this is true then the package_data won't be included in the
    # dist. Use MANIFEST.in for this
    include_package_data=True,

    # pymodules is for single file standalone modules not part of the
    # package
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],

    entry_points={
        'console_scripts' : [
            'jubeo=jubeo.cli:cli',
        ]
    },

    install_requires=base_requirements,

    # SNIPPET: example of using extra requirements
    # extras_require={
    #     'extras' : example_extra_requirements
    #     'all' : all_requirements,
    # }

    # include your extra requirement sets here
    extras_require={
        'all' : all_requirements,
    }
)




# IDEA: make a __version_info__ attribute that follows "growth versioning semantics"
# SNIPPET: for that idea
# breakage = 0
# regression = 0
# growth = 0
# prerelease_type = "a"
# prerelease_number = 0
# dev = 0
# postrelease = 0

# __version_info__ = (breakage,
#                     regression,
#                     growth,
#                     prerelease_type,
#                     prerelease_number,
#                     dev,
#                     postrelease,)

# __version__ = f"{breakage}.{regression}.{growth}{prerelease_type}{prerelease_number}.dev{dev}.post{postrelease}"

__version__ = '0.0.3a0.dev0'



import shutil
import os
import os.path as osp
from pathlib import Path

import click

from hyperlink import URL
import toml

from .main import (
    base_repo_path,
    config_template_path,
    retrieve_upstream,
    init_project,
    load_proj_config,
    update_project,
)

CONFIG_DIR = "$XDG_CONFIG_HOME/jubeo"
"""The system-wide configuration directory, including configuring the
jubeo tool itself"""

JUBEO_CONF_NAME = "jubeo_conf.toml"
"""Configures the behavior of the jubeo tool."""

def load_config():

    config_path = Path(osp.expandvars(CONFIG_DIR)) / JUBEO_CONF_NAME
    config = toml.load(config_path)

    return config

def _init_cache():

    config = load_config()
    cache_path = Path(osp.expandvars(osp.expanduser(config['cache']['path'])))

    os.makedirs(cache_path,
                exist_ok=True)

    # make the sub directories needed
    os.makedirs(cache_path / 'specs',
                exist_ok=True)
    os.makedirs(cache_path / 'modules',
                exist_ok=True)

@click.command()
@click.option("--force/--no-force", default=False)
def init_config(force):
    """Initialize the system-wide configuration directory."""

    config_dir_path = config_template_path()

    conf_dir = Path(osp.expandvars(CONFIG_DIR))

    if osp.exists(conf_dir) and force:
        shutil.rmtree(conf_dir)

    shutil.copytree(
        config_dir_path,
        conf_dir,
        dirs_exist_ok=force,
        # ignore the mod def which is needed for pkgutils
        ignore=shutil.ignore_patterns("*~", "__init__.py"),
    )

    _init_cache()



def _clean_cache():
    """Remove cached files."""

    config = load_config()

    try:
        shutil.rmtree(osp.expanduser(osp.expandvars(config['cache']['path'])))
    except FileNotFoundError:
        _init_cache()
    else:
        _init_cache()



@click.command()
def clean_cache():
    _clean_cache()


@click.command()
@click.option("--force/--no-force", default=False)
@click.option("--taskset-name", default="tasks")
@click.option('--upstream', type=str, default=None)
@click.option("--base-repo", is_flag=True, default=False,
              help="Choose to initialize with the base repo in jubeo source code."
              "Otherwise use the jubeo project config if it exists.")
@click.argument('project-dir', type=click.Path(exists=True))
def init(force, taskset_name, upstream, base_repo, project_dir):
    """Initialize a project task-set.

    If the '--upstream' option is not given will look to whether the
    '--base-repo' flag is on.

    If the '--base-repo' flag is on, jubeo will always initialize this
    directory with the 'base_repo' configuration which is included in
    the jubeo source code. Otherwise if a '.jubeo/jubeo.toml' file is
    present in the project this will be used. If it doesn't exist then
    the 'base_repo' will be used.

    """

    _clean_cache()

    project_dir = Path(project_dir)

    config = load_config()

    cache_path = Path(osp.expanduser(osp.expandvars(config['cache']['path'])))

    # see if the project has a config, and load it if there is
    try:
        proj_config = load_proj_config(project_dir)
    except FileNotFoundError:
        proj_config = None

    # get the upstream repo we are retrieving from, downloading if
    # necessary
    if upstream is None:

        # if base_repo option not on and this project has a config
        # load this config

        if (not base_repo and
            proj_config is not None):

            # read the upstream URL from the config
            upstream_url = URL.from_text(osp.expandvars(osp.expanduser(
                proj_config['upstream_url'])))


        # fall back to base_repo if everything else failed
        if base_repo:
            upstream_url = URL.from_text(str(base_repo_path()))


    else:
        upstream_url = URL.from_text(upstream)


    # then we can simply copy from this downloaded repo
    source_repo_path = Path(retrieve_upstream(
        upstream_url,
        cache_path,
        proj_config = proj_config,
    ))

    init_project(
        source_repo_path,
        project_dir,
        force=force,
    )



@click.command()
@click.argument('project-dir', type=click.Path(exists=True))
def update(project_dir):
    """Update a projects task-set."""

    # we don't do a fancy caching strategy and redownload every time
    # now, so we clean everytime now
    _clean_cache()

    project_dir = Path(project_dir)

    # read the system-wide configuration file to find location of the
    # cache directory
    config = load_config()

    # load the project config
    proj_config = load_proj_config(project_dir)

    cache_path = Path(osp.expanduser(osp.expandvars(config['cache']['path'])))

    # get the upstream URL
    url_str = osp.expandvars(proj_config['upstream_url'])
    url = URL.from_text(url_str)

    module_url = URL.from_text(
        osp.expandvars(osp.expanduser(proj_config['modules']['source_url'])))

    print(f"Using upstream: {url}")
    print(f"Retrieving modules from: {module_url}")

    # retrieve the upstream repo and return the exact file location of
    # the jubeo bundle
    source_repo_path = retrieve_upstream(
        url,
        cache_path,
        proj_config,
    )

    ## update the project
    update_project(
        source_repo_path,
        project_dir,
        proj_config,
        force=True,
    )


@click.group()
def cli():
    pass

cli.add_command(update)
cli.add_command(init)
cli.add_command(clean_cache)
cli.add_command(init_config)

if __name__ == "__main__":

    cli()
